# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const
from neutron_lib.plugins.ml2 import api as driver_api
from neutron_lib.services.trunk import constants as trunk_consts
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from neutron.db import segments_db

from networking_arista.common import constants as a_const
from networking_arista.common import db_lib
from networking_arista.ml2 import arista_sync
from networking_arista.ml2 import arista_trunk
from networking_arista.ml2.rpc.arista_eapi import AristaRPCWrapperEapi

# When used as a Neutron plugin, neutron-lib imports this code. However earlier
# neutron-lib has already imported 'multiprocessing'. This means the python
# module cache (sys.modules) contains a version of 'multiprocessing' where
# select.poll() exists.
#
# Further down we import arista_sync, which spawns a greenthread. This
# greenthread then uses a green version of 'multiprocessing' where
# select.poll() has been removed.
#
# Doing here multiprocessing.Queue.put() and in the greenthread
# multiprocessing.Queue.get(timeout=...) leads to:
# AttributeError: module 'select' has no attribute 'poll'
#
# We can't do eventlet.monkey_patch() early enough (before the first
# 'mutiprocessing' import) as we would have to do it in neutron-lib and it's
# forbidden, see https://review.opendev.org/#/c/333017/
#
# The solution is to let the python module cache here forget the already
# imported 'multiprocessing' and re-import a green one. Here again
# eventlet.monkey_patch() doesn't seem to help as it doesn't seem to touch
# 'multiprocessing'. Thus we use eventlet.import_patched() instead:
import eventlet
import sys
modules_to_forget = []
for imported_module_name in sys.modules:
    if imported_module_name.startswith('multiprocessing'):
        modules_to_forget.append(imported_module_name)
for module_to_forget in modules_to_forget:
    del sys.modules[module_to_forget]

for module_to_forget in modules_to_forget:
    try:
        eventlet.import_patched(module_to_forget)
    except ImportError:
        pass

# import a green 'multiprocessing':
multiprocessing = eventlet.import_patched('multiprocessing')
from networking_arista.ml2 import arista_sync  # noqa: E402

LOG = logging.getLogger(__name__)
cfg.CONF.import_group('ml2_arista', 'networking_arista.common.config')


def log_context(function, context):
    pretty_context = json.dumps(context, sort_keys=True, indent=4)
    LOG.debug(function)
    LOG.debug(pretty_context)


class MechResource(object):
    """Container class for passing data to sync worker"""

    def __init__(self, id, resource_type, action, related_resources=None):
        self.id = id
        self.resource_type = resource_type
        self.action = action
        self.related_resources = related_resources or list()

    def __str__(self):
        return "%s %s ID: %s" % (self.action, self.resource_type, self.id)


class AristaDriver(driver_api.MechanismDriver):
    """Ml2 Mechanism driver for Arista networking hardware.

    Remembers all networks and VMs that are provisioned on Arista Hardware.
    Does not send network provisioning request if the network has already been
    provisioned before for the given port.
    """
    def __init__(self):
        confg = cfg.CONF.ml2_arista
        self.managed_physnets = confg['managed_physnets']
        self.manage_fabric = confg['manage_fabric']
        self.eapi = AristaRPCWrapperEapi()
        self.mlag_pairs = dict()

        self.provision_queue = multiprocessing.Queue()
        self.trunk_driver = None

    def initialize(self):
        self.mlag_pairs = db_lib.get_mlag_physnets()
        self.trunk_driver = arista_trunk.AristaTrunkDriver.create()

    def get_workers(self):
        return [arista_sync.AristaSyncWorker(self.provision_queue)]

    def create_network(self, network, segments):
        """Enqueue network create"""
        tenant_id = network['project_id']
        action = a_const.CREATE if tenant_id else a_const.FULL_SYNC
        n_res = MechResource(network['id'], a_const.NETWORK_RESOURCE, action)
        n_res.related_resources.append((a_const.TENANT_RESOURCE, tenant_id))
        for segment in segments:
            n_res.related_resources.append(
                (a_const.SEGMENT_RESOURCE, segment['id']))
        self.provision_queue.put(n_res)

    def delete_network(self, network, segments):
        """Enqueue network delete"""
        tenant_id = network['project_id']
        action = a_const.DELETE if tenant_id else a_const.FULL_SYNC
        n_res = MechResource(network['id'], a_const.NETWORK_RESOURCE, action)

        # Delete tenant if this was the last tenant resource
        if not db_lib.tenant_provisioned(tenant_id):
            n_res.related_resources.append(
                (a_const.TENANT_RESOURCE, tenant_id))

        for segment in segments:
            n_res.related_resources.append(
                (a_const.SEGMENT_RESOURCE, segment['id']))
        self.provision_queue.put(n_res)

    def delete_segment(self, segment):
        """Enqueue segment delete"""
        s_res = MechResource(segment['id'], a_const.SEGMENT_RESOURCE,
                             a_const.DELETE)
        self.provision_queue.put(s_res)

    def get_instance_type(self, port):
        """Determine the port type based on device owner and vnic type"""
        if port[portbindings.VNIC_TYPE] == portbindings.VNIC_BAREMETAL:
            return a_const.BAREMETAL_RESOURCE
        owner_to_type = {
            n_const.DEVICE_OWNER_DHCP: a_const.DHCP_RESOURCE,
            n_const.DEVICE_OWNER_DVR_INTERFACE: a_const.ROUTER_RESOURCE,
            n_const.DEVICE_OWNER_ROUTER_INTF: a_const.ROUTER_RESOURCE,
            n_const.DEVICE_OWNER_ROUTER_HA_INTF: a_const.ROUTER_RESOURCE,
            trunk_consts.TRUNK_SUBPORT_OWNER: a_const.VM_RESOURCE}
        if port['device_owner'] in owner_to_type.keys():
            return owner_to_type[port['device_owner']]
        elif port['device_owner'].startswith(
                n_const.DEVICE_OWNER_COMPUTE_PREFIX):
            return a_const.VM_RESOURCE
        return None

    def _get_binding_keys(self, port, host):
        """Get binding keys from the port binding"""
        binding_keys = list()
        switch_binding = port[portbindings.PROFILE].get(
            'local_link_information', None)
        if switch_binding:
            for binding in switch_binding:
                switch_id = binding.get('switch_id')
                port_id = binding.get('port_id')
                binding_keys.append((port['id'], (switch_id, port_id)))
        else:
            binding_keys.append((port['id'], host))
        return binding_keys

    def create_port_binding(self, port, host):
        """Enqueue port binding create"""
        tenant_id = port['project_id']
        instance_type = self.get_instance_type(port)
        if not instance_type:
            return
        port_type = instance_type + a_const.PORT_SUFFIX
        action = a_const.CREATE if tenant_id else a_const.FULL_SYNC
        related_resources = list()
        related_resources.append((a_const.TENANT_RESOURCE, tenant_id))
        related_resources.append((instance_type, port['device_id']))
        related_resources.append((port_type, port['id']))
        for pb_key in self._get_binding_keys(port, host):
            pb_res = MechResource(pb_key, a_const.PORT_BINDING_RESOURCE,
                                  action, related_resources=related_resources)
            self.provision_queue.put(pb_res)

    def delete_port_binding(self, port, host):
        """Enqueue port binding delete"""
        tenant_id = port['project_id']
        instance_type = self.get_instance_type(port)
        if not instance_type:
            return
        port_type = instance_type + a_const.PORT_SUFFIX
        action = a_const.DELETE if tenant_id else a_const.FULL_SYNC
        related_resources = list()

        # Delete tenant if this was the last tenant resource
        if not db_lib.tenant_provisioned(tenant_id):
            related_resources.append((a_const.TENANT_RESOURCE, tenant_id))

        # Delete instance if this was the last instance port
        if not db_lib.instance_provisioned(port['device_id']):
            related_resources.append((instance_type, port['device_id']))

        # Delete port if this was the last port binding
        if not db_lib.port_provisioned(port['id']):
            related_resources.append((port_type, port['id']))

        for pb_key in self._get_binding_keys(port, host):
            pb_res = MechResource(pb_key, a_const.PORT_BINDING_RESOURCE,
                                  action, related_resources=related_resources)
            self.provision_queue.put(pb_res)

    def create_network_postcommit(self, context):
        """Provision the network on CVX"""
        network = context.current

        log_context("create_network_postcommit: network", network)

        segments = context.network_segments
        self.create_network(network, segments)

    def update_network_postcommit(self, context):
        """Send network updates to CVX:

        - Update the network name
        - Add new segments
        """
        network = context.current
        orig_network = context.original

        log_context("update_network_postcommit: network", network)
        log_context("update_network_postcommit: orig", orig_network)

        segments = context.network_segments

        # New segments may have been added
        self.create_network(network, segments)

    def delete_network_postcommit(self, context):
        """Delete the network from CVX"""
        network = context.current

        log_context("delete_network_postcommit: network", network)

        segments = context.network_segments
        self.delete_network(network, segments)

    def update_port_postcommit(self, context):
        """Send port updates to CVX

        This method is also responsible for the initial creation of ports
        as we wait until after a port is bound to send the port data to CVX
        """
        port = context.current
        orig_port = context.original
        network = context.network.current

        log_context("update_port_postcommit: port", port)
        log_context("update_port_postcommit: orig", orig_port)

        # Device id can change without a port going DOWN, but the new device
        # id may not be supported
        if orig_port and port['device_id'] != orig_port['device_id']:
            self.delete_port_binding(orig_port, context.original_host)

        if context.status in [n_const.PORT_STATUS_ACTIVE,
                              n_const.PORT_STATUS_BUILD]:
            if context.binding_levels:
                segments = [
                    level['bound_segment'] for level in context.binding_levels]
                self.create_network(network, segments)
            self.create_port_binding(port, context.host)
        else:
            if (context.original_host and
                    context.status != context.original_status):
                self.delete_port_binding(orig_port, context.original_host)
                self._try_to_release_dynamic_segment(context, migration=True)

    def delete_port_postcommit(self, context):
        """Delete the port from CVX"""
        port = context.current

        log_context("delete_port_postcommit: port", port)

        self.delete_port_binding(port, context.host)
        self._try_to_release_dynamic_segment(context)

    def _bind_baremetal_port(self, context, segment):
        """Bind the baremetal port to the segment"""
        port = context.current
        vif_details = {
            portbindings.VIF_DETAILS_VLAN: str(
                segment[driver_api.SEGMENTATION_ID])
        }
        context.set_binding(segment[driver_api.ID],
                            portbindings.VIF_TYPE_OTHER,
                            vif_details,
                            n_const.ACTIVE)
        LOG.debug("AristaDriver: bound port info- port ID %(id)s "
                  "on network %(network)s",
                  {'id': port['id'],
                   'network': context.network.current['id']})
        if port.get('trunk_details'):
            self.trunk_driver.bind_port(port)
        return True

    def _get_physnet(self, context):
        """Find the appropriate physnet for the host

        - Baremetal ports' physnet is determined by looking at the
          local_link_information contained in the binding profile
        - Other ports' physnet is determined by looking for the host in the
          topology
        """
        port = context.current
        physnet = None
        if (port.get(portbindings.VNIC_TYPE) == portbindings.VNIC_BAREMETAL):
            physnet = self.eapi.get_baremetal_physnet(context)
        else:
            physnet = self.eapi.get_host_physnet(context)
        # If the switch is part of an mlag pair, the physnet is called
        # peer1_peer2
        physnet = self.mlag_pairs.get(physnet, physnet)
        return physnet

    def _bind_fabric(self, context, segment):
        """Allocate dynamic segments for the port

        Segment physnets are based on the switch to which the host is
        connected.
        """
        port_id = context.current['id']
        physnet = self._get_physnet(context)
        if not physnet:
            LOG.debug("bind_port for port %(port)s: no physical_network "
                      "found", {'port': port_id})
            return False

        with lockutils.lock(physnet, external=True):
            context.allocate_dynamic_segment(
                {'network_id': context.network.current['id'],
                 'network_type': n_const.TYPE_VLAN,
                 'physical_network': physnet})
        next_segment = segments_db.get_dynamic_segment(
            context._plugin_context, context.network.current['id'],
            physical_network=physnet)
        LOG.debug("bind_port for port %(port)s: "
                  "current_segment=%(current_seg)s, "
                  "next_segment=%(next_seg)s",
                  {'port': port_id, 'current_seg': segment,
                   'next_seg': next_segment})
        context.continue_binding(segment['id'], [next_segment])
        return True

    def bind_port(self, context):
        """Bind port to a network segment.

        Provisioning request to Arista Hardware to plug a host
        into appropriate network is done when the port is created
        this simply tells the ML2 Plugin that we are binding the port
        """
        port = context.current

        log_context("bind_port: port", port)

        for segment in context.segments_to_bind:
            physnet = segment.get(driver_api.PHYSICAL_NETWORK)
            segment_type = segment[driver_api.NETWORK_TYPE]
            if not physnet:
                if (segment_type == n_const.TYPE_VXLAN and self.manage_fabric):
                    if self._bind_fabric(context, segment):
                        continue
            elif (port.get(portbindings.VNIC_TYPE)
                    == portbindings.VNIC_BAREMETAL):
                if (not self.managed_physnets or
                        physnet in self.managed_physnets):
                    if self._bind_baremetal_port(context, segment):
                        continue
            LOG.debug("Arista mech driver unable to bind port %(port)s to "
                      "%(seg_type)s segment on physical_network %(physnet)s",
                      {'port': port.get('id'), 'seg_type': segment_type,
                       'physnet': physnet})

    def _try_to_release_dynamic_segment(self, context, migration=False):
        """Release dynamic segment if necessary

        If this port was the last port using a segment and the segment was
        allocated by this driver, it should be released
        """
        if migration:
            binding_levels = context.original_binding_levels
        else:
            binding_levels = context.binding_levels
        LOG.debug("_try_to_release_dynamic_segment: "
                  "binding_levels=%(bl)s", {'bl': binding_levels})
        if not binding_levels:
            return

        for prior_level, binding in enumerate(binding_levels[1:]):
            allocating_driver = binding_levels[prior_level].get(
                driver_api.BOUND_DRIVER)
            if allocating_driver != a_const.MECHANISM_DRV_NAME:
                continue
            bound_segment = binding.get(driver_api.BOUND_SEGMENT, {})
            segment_id = bound_segment.get('id')
            if not db_lib.segment_is_dynamic(segment_id):
                continue
            if not db_lib.segment_bound(segment_id):
                context.release_dynamic_segment(segment_id)
                self.delete_segment(bound_segment)
                LOG.debug("Released dynamic segment %(seg)s allocated "
                          "by %(drv)s", {'seg': segment_id,
                                         'drv': allocating_driver})



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Exceptions used by Arista ML2 Mechanism Driver."""

from neutron_lib import exceptions

from networking_arista._i18n import _


class AristaRpcError(exceptions.NeutronException):
    message = _('%(msg)s')


class AristaConfigError(exceptions.NeutronException):
    message = _('%(msg)s')


class AristaServicePluginRpcError(exceptions.NeutronException):
    message = _('%(msg)s')


class AristaServicePluginConfigError(exceptions.NeutronException):
    message = _('%(msg)s')


class AristaServicePluginInvalidCommand(exceptions.NeutronException):
    message = _('%(msg)s')


class VlanUnavailable(exceptions.NeutronException):
    """An exception indicating VLAN creation failed because it's not available.

    A specialization of the NeutronException indicating network creation failed
    because a specified VLAN is unavailable on the physical network.

    :param vlan_id: The VLAN ID.
    :param physical_network: The physical network.
    """
    message = _("Unable to create the network. "
                "The VLAN %(vlan_id)s on physical network "
                "%(physical_network)s is not available.")



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib.plugins.ml2 import api as driver_api


class TestFabricDriver(driver_api.MechanismDriver):

    def initialize(self):
        pass

    def bind_port(self, context):
        """Bind port to a network segment."""
        port = context.current
        for segment in context.segments_to_bind:
            physnet = segment.get(driver_api.PHYSICAL_NETWORK)
            if (not physnet and
                    segment[driver_api.NETWORK_TYPE] == constants.TYPE_VXLAN):
                physnet_map = {'host1': 'physnet1',
                               'host2': 'physnet2'}
                physnet = physnet_map.get(port[portbindings.HOST_ID],
                                          'other_physnet')
                next_segment = context.allocate_dynamic_segment(
                    {'network_id': context.network.current['id'],
                     'network_type': constants.TYPE_VLAN,
                     'physical_network': physnet})
                context.continue_binding(segment['id'], [next_segment])
                return True



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg

from neutron.tests import base
from neutron.tests.unit import testlib_api
from neutron_lib.api.definitions import portbindings

from networking_arista.common import exceptions as arista_exc
from networking_arista.ml2.rpc import arista_eapi
from networking_arista.tests.unit import utils


def setup_valid_config():
    utils.setup_arista_wrapper_config(cfg)


class AristaRPCWrapperInvalidConfigTestCase(base.BaseTestCase):
    """Negative test cases to test the Arista Driver configuration."""

    def setUp(self):
        super(AristaRPCWrapperInvalidConfigTestCase, self).setUp()
        self.setup_invalid_config()  # Invalid config, required options not set

    def setup_invalid_config(self):
        utils.setup_arista_wrapper_config(cfg, host='', user='')

    def test_raises_exception_on_wrong_configuration(self):
        self.assertRaises(arista_exc.AristaConfigError,
                          arista_eapi.AristaRPCWrapperEapi)


class NegativeRPCWrapperTestCase(testlib_api.SqlTestCase):
    """Negative test cases to test the RPC between Arista Driver and EOS."""

    def setUp(self):
        super(NegativeRPCWrapperTestCase, self).setUp()
        setup_valid_config()

    def test_exception_is_raised_on_json_server_error(self):
        drv = arista_eapi.AristaRPCWrapperEapi()

        drv.api_request = mock.MagicMock(
            side_effect=Exception('server error')
        )
        with mock.patch.object(arista_eapi.LOG, 'error') as log_err:
            self.assertRaises(arista_exc.AristaRpcError,
                              drv._run_openstack_cmds, [])
            log_err.assert_called_once_with(mock.ANY)


class GetPhysnetTestCase(base.BaseTestCase):
    """Test cases to validate parsing of topology output to find physnets"""

    def setUp(self):
        super(GetPhysnetTestCase, self).setUp()
        setup_valid_config()

    def _test_get_host_physnet(self, nova_fqdn, topo_host_fqdn,
                               topo_switch_fqdn, bridge_map_fqdn, use_fqdn,
                               use_fqdn_physnet):
        cfg.CONF.set_override('use_fqdn', use_fqdn, "ml2_arista")
        cfg.CONF.set_override('use_fqdn_physnet', use_fqdn_physnet,
                              "ml2_arista")
        context = mock.MagicMock()
        nova_host1 = 'host1.full.name' if nova_fqdn else 'host1'
        nova_host2 = 'host2.full.name' if nova_fqdn else 'host2'
        topo_host1 = 'host1.full.name' if topo_host_fqdn else 'host1'
        topo_host2 = 'host2.full.name' if topo_host_fqdn else 'host2'
        topo_switch1 = 'switch1.full.name' if topo_switch_fqdn else 'switch1'
        topo_switch2 = 'switch2.full.name' if topo_switch_fqdn else 'switch2'
        bridge_map_switch1 = ('switch1.full.name' if bridge_map_fqdn
                              else 'switch1')
        bridge_map_switch2 = ('switch2.full.name' if bridge_map_fqdn
                              else 'switch2')
        context.host = nova_host1
        topology = [{'neighbors':
                     {'%s-et1' % topo_host1:
                      {'fromPort':
                       {'name': 'et1',
                        'hostname': topo_host1,
                        'hostid': '00:00:00:00:00:00'},
                       'toPort': [
                           {'name': 'Ethernet1',
                            'hostname': topo_switch1,
                            'hostid': '00:00:00:00:00:01'}]},
                      '%s-et1' % topo_host2:
                      {'fromPort':
                       {'name': 'et1',
                        'hostname': topo_host2,
                        'hostid': '00:00:00:00:00:02'},
                       'toPort': [
                           {'name': 'Ethernet1',
                            'hostname': topo_switch2,
                            'hostid': '00:00:00:00:00:03'}]}}}]
        drv = arista_eapi.AristaRPCWrapperEapi()
        drv._run_eos_cmds = mock.MagicMock()
        drv._run_eos_cmds.return_value = topology
        self.assertEqual(drv.get_host_physnet(context), bridge_map_switch1)
        context.host = nova_host2
        self.assertEqual(drv.get_host_physnet(context), bridge_map_switch2)

    def _test_get_baremetal_physnet(self, topo_switch_fqdn, bridge_map_fqdn,
                                    use_fqdn_physnet):
        cfg.CONF.set_override('use_fqdn_physnet', use_fqdn_physnet,
                              "ml2_arista")
        context = mock.MagicMock()
        topo_switch1 = 'switch1.full.name' if topo_switch_fqdn else 'switch1'
        topo_switch2 = 'switch2.full.name' if topo_switch_fqdn else 'switch2'
        bridge_map_switch1 = ('switch1.full.name' if bridge_map_fqdn
                              else 'switch1')
        bridge_map_switch2 = ('switch2.full.name' if bridge_map_fqdn
                              else 'switch2')
        context.host = 'host1'
        context.current = {portbindings.PROFILE: {
            'local_link_information': [{'switch_id': '00:00:00:00:00:00'}]}}
        topology = [{'hosts':
                     {'00:00:00:00:00:00': {'name': '00:00:00:00:00:00',
                                            'hostname': topo_switch1},
                      '00:00:00:00:00:01': {'name': '00:00:00:00:00:01',
                                            'hostname': topo_switch2}}}]
        drv = arista_eapi.AristaRPCWrapperEapi()
        drv._run_eos_cmds = mock.MagicMock()
        drv._run_eos_cmds.return_value = topology
        self.assertEqual(drv.get_baremetal_physnet(context),
                         bridge_map_switch1)
        context.host = 'host2'
        context.current = {portbindings.PROFILE: {
            'local_link_information': [{'switch_id': '00:00:00:00:00:01'}]}}
        self.assertEqual(drv.get_baremetal_physnet(context),
                         bridge_map_switch2)

    def test_get_host_physnet(self):
        for nova_fqdn in (True, False):
            for topo_host_fqdn in (True, False):
                for topo_switch_fqdn in (True, False):
                    for bridge_map_fqdn in (True, False):
                        if bridge_map_fqdn and not topo_switch_fqdn:
                            # Topology has less info than bridge map.
                            # This isn't supported
                            continue
                        use_fqdn = True
                        if nova_fqdn and not topo_host_fqdn:
                            use_fqdn = False
                        use_fqdn_physnet = True
                        if topo_switch_fqdn and not bridge_map_fqdn:
                            use_fqdn_physnet = False
                        self._test_get_host_physnet(nova_fqdn,
                                                    topo_host_fqdn,
                                                    topo_switch_fqdn,
                                                    bridge_map_fqdn,
                                                    use_fqdn,
                                                    use_fqdn_physnet)

    def test_get_baremetal_physnet(self):
        for topo_switch_fqdn in (True, False):
            for bridge_map_fqdn in (True, False):
                if bridge_map_fqdn and not topo_switch_fqdn:
                    # Topology has less info than bridge map.
                    # This isn't supported.
                    continue
                use_fqdn_physnet = True
                if topo_switch_fqdn and not bridge_map_fqdn:
                    use_fqdn_physnet = False
                self._test_get_baremetal_physnet(topo_switch_fqdn,
                                                 bridge_map_fqdn,
                                                 use_fqdn_physnet)



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.db import api as db_api
from oslo_log import log
from six import moves

from neutron.objects import network as network_obj
from neutron.objects.plugins.ml2 import vlanallocation as vlanalloc

from networking_arista._i18n import _LI
from networking_arista.common.constants import EOS_UNREACHABLE_MSG
from networking_arista.common import exceptions as arista_exc

LOG = log.getLogger(__name__)


class VlanSyncService(object):
    """Sync vlan assignment from CVX into the OpenStack db."""

    def __init__(self, rpc_wrapper):
        self._rpc = rpc_wrapper
        self._force_sync = True
        self._vlan_assignment_uuid = None
        self._assigned_vlans = dict()

    def force_sync(self):
        self._force_sync = True

    def _parse_vlan_ranges(self, vlan_pool, return_as_ranges=False):
        vlan_ids = set()
        if return_as_ranges:
            vlan_ids = list()
        if not vlan_pool:
            return vlan_ids
        vlan_ranges = vlan_pool.split(',')
        for vlan_range in vlan_ranges:
            endpoints = vlan_range.split('-')
            if len(endpoints) == 2:
                vlan_min = int(endpoints[0])
                vlan_max = int(endpoints[1])
                if return_as_ranges:
                    vlan_ids.append((vlan_min, vlan_max))
                else:
                    vlan_ids |= set(moves.range(vlan_min, vlan_max + 1))
            elif len(endpoints) == 1:
                single_vlan = int(endpoints[0])
                if return_as_ranges:
                    vlan_ids.append((single_vlan, single_vlan))
                else:
                    vlan_ids.add(single_vlan)
        return vlan_ids

    def get_network_vlan_ranges(self):
        return self._assigned_vlans

    def _sync_required(self):
        try:
            if not self._force_sync and self._region_in_sync():
                LOG.info(_LI('VLANs are in sync!'))
                return False
        except arista_exc.AristaRpcError:
            LOG.warning(EOS_UNREACHABLE_MSG)
            self._force_sync = True
        return True

    def _region_in_sync(self):
        eos_vlan_assignment_uuid = self._rpc.get_vlan_assignment_uuid()
        return (self._vlan_assignment_uuid and
                (self._vlan_assignment_uuid['uuid'] ==
                 eos_vlan_assignment_uuid['uuid']))

    def _set_vlan_assignment_uuid(self):
        try:
            self._vlan_assignment_uuid = self._rpc.get_vlan_assignment_uuid()
        except arista_exc.AristaRpcError:
            self._force_sync = True

    def do_synchronize(self):
        if not self._sync_required():
            return

        self.synchronize()
        self._set_vlan_assignment_uuid()

    def synchronize(self):
        LOG.info(_LI('Syncing VLANs with EOS'))
        try:
            self._rpc.check_vlan_type_driver_commands()
            vlan_pool = self._rpc.get_vlan_allocation()
        except arista_exc.AristaRpcError:
            LOG.warning(EOS_UNREACHABLE_MSG)
            self._force_sync = True
            return

        LOG.debug('vlan_pool %(vlan)s', {'vlan': vlan_pool})
        self._assigned_vlans = {
            'default': self._parse_vlan_ranges(vlan_pool['assignedVlans']),
        }
        cvx_available_vlans = frozenset(
            self._parse_vlan_ranges(vlan_pool['availableVlans']))
        cvx_used_vlans = frozenset(
            self._parse_vlan_ranges(vlan_pool['allocatedVlans']))
        # Force vlan sync if assignedVlans is empty or availableVlans and
        # allocatedVlans both are empty in the vlan_pool
        if not(self._assigned_vlans['default'] and
               (cvx_available_vlans or cvx_used_vlans)):
            LOG.info(_LI('Force sync, vlan pool is empty'))
            self.force_sync()
        else:
            self._force_sync = False

        allocated_vlans = {}
        ctx = context.get_admin_context()
        with db_api.CONTEXT_READER.using(ctx):
            for physical_network in self._assigned_vlans:
                filter = {
                    'network_type': n_const.TYPE_VLAN,
                    'physical_network': physical_network,
                }
                objs = network_obj.NetworkSegment.get_objects(ctx, **filter)
                allocated_vlans.update(
                    {physical_network: [obj.segmentation_id for obj in objs]})
            LOG.debug('allocated vlans %(vlan)s', {'vlan': allocated_vlans})

        with db_api.CONTEXT_WRITER.using(ctx):
            physnets = vlanalloc.VlanAllocation.get_physical_networks(ctx)
            physnets_unconfigured = physnets - set(self._assigned_vlans)
            if physnets_unconfigured:
                vlanalloc.VlanAllocation.delete_physical_networks(
                    ctx, physnets_unconfigured)

            allocations = collections.defaultdict(list)
            for alloc in vlanalloc.VlanAllocation.get_objects(ctx):
                allocations[alloc.physical_network].append(alloc)

            for physical_network, vlan_ranges in self._assigned_vlans.items():
                if physical_network in allocations:
                    for alloc in allocations[physical_network]:
                        try:
                            vlan_ranges.remove(alloc.vlan_id)
                        except KeyError:
                            alloc.delete()
                vlanalloc.VlanAllocation.bulk_create(ctx, physical_network,
                                                     vlan_ranges)
                LOG.debug('vlan_ranges: %(vlan)s', {'vlan': vlan_ranges})
                for vlan_id in vlan_ranges:
                    allocated = (vlan_id not in cvx_available_vlans and
                                 (vlan_id in cvx_used_vlans or vlan_id in
                                  allocated_vlans[physical_network]))
                    LOG.debug('Updating %(phys)s %(vlan)s %(alloc)s',
                              {'phys': physical_network, 'vlan': vlan_id,
                               'alloc': allocated})
                    vlanalloc.VlanAllocation.update_objects(
                        ctx,
                        values={'allocated': allocated,
                                'vlan_id': vlan_id,
                                'physical_network': physical_network},
                        physical_network=physical_network,
                        vlan_id=vlan_id)



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import operator
import socket

import mock
from mock import patch
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_utils import importutils

from neutron.tests.unit import testlib_api

from networking_arista.ml2.rpc import arista_json
from networking_arista.tests.unit import utils


BASE_RPC = "networking_arista.ml2.rpc.arista_json.AristaRPCWrapperJSON."
JSON_SEND_FUNC = BASE_RPC + "send_api_request"
RAND_FUNC = BASE_RPC + "_get_random_name"
DB_LIB_MODULE = 'networking_arista.ml2.rpc.arista_json.db_lib'


def setup_valid_config():
    utils.setup_arista_wrapper_config(cfg)


class _UnorderedDictList(list):
    def __init__(self, iterable='', sort_key=None):
        super(_UnorderedDictList, self).__init__(iterable)
        try:
            (self[0] or {})[sort_key]
            self.sort_key = sort_key
        except (IndexError, KeyError):
            self.sort_key = None

    def __eq__(self, other):
        if isinstance(other, list) and self.sort_key:
            key = operator.itemgetter(self.sort_key)
            return sorted(self, key=key) == sorted(other, key=key)
        else:
            return super(_UnorderedDictList, self).__eq__(other)


class TestAristaJSONRPCWrapper(testlib_api.SqlTestCase):
    def setUp(self):
        super(TestAristaJSONRPCWrapper, self).setUp()
        plugin_klass = importutils.import_class(
            "neutron.db.db_base_plugin_v2.NeutronDbPluginV2")
        directory.add_plugin(plugin_constants.CORE, plugin_klass())
        setup_valid_config()
        self.drv = arista_json.AristaRPCWrapperJSON()
        self.drv._server_ip = "10.11.12.13"
        self.region = 'RegionOne'

    def _verify_send_api_request_call(self, mock_send_api_req, calls,
                                      unordered_dict_list=False):
        if unordered_dict_list:
            wrapper = functools.partial(_UnorderedDictList, sort_key='id')
        else:
            wrapper = lambda x: x

        expected_calls = [
            mock.call(c[0], c[1], *(wrapper(d) for d in c[2:])) for c in calls
        ]

        mock_send_api_req.assert_has_calls(expected_calls, any_order=True)

    @patch(JSON_SEND_FUNC)
    def test_register_with_eos(self, mock_send_api_req):
        self.drv.register_with_eos()
        calls = [
            ('region/RegionOne', 'PUT',
             [{'name': 'RegionOne', 'syncInterval': 1}])
        ]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    def _get_random_name(self):
        return 'thisWillBeRandomInProd'

    @patch(JSON_SEND_FUNC)
    @patch(RAND_FUNC, _get_random_name)
    def test_sync_start(self, mock_send_api_req):
        mock_send_api_req.side_effect = [
            [{'name': 'RegionOne', 'syncStatus': '',
              'syncInterval': self.drv.sync_interval}],
            [{}]
        ]
        assert self.drv.sync_start()
        calls = [
            ('region/RegionOne/sync', 'POST',
             {'requester': socket.gethostname().split('.')[0],
              'requestId': self._get_random_name()})
        ]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    @patch('requests.Response')
    def test_sync_start_exception(self, mock_response):
        mock_response.ok.return_value = False
        self.assertFalse(self.drv.sync_start())

    @patch(JSON_SEND_FUNC)
    def test_sync_start_no_region(self, mock_send_api_req):
        mock_send_api_req.return_value = {}
        self.assertFalse(self.drv.sync_start())
        calls = [
            ('region/RegionOne', 'GET'),
            ('region/', 'POST', [{'name': 'RegionOne'}])
        ]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    def _get_region(self, region):
        return {'name': region, 'syncStatus': 'syncTimedout',
                'syncInterval': self.sync_interval}

    @patch('requests.post')
    @patch(BASE_RPC + 'get_region', _get_region)
    @patch(BASE_RPC + '_get_eos_master', lambda _: 'cvx')
    @patch(RAND_FUNC, _get_random_name)
    def test_sync_start_after_failure(self, mock_post):
        self.drv.current_sync_name = 'bad-sync-id'
        self.assertTrue(self.drv.sync_start())
        expected_header = {'Content-Type': 'application/json',
                           'Accept': 'application/json',
                           'X-Sync-ID': None}
        mock_post.assert_called_once_with(mock.ANY,
                                          data=mock.ANY,
                                          timeout=mock.ANY,
                                          verify=mock.ANY,
                                          headers=expected_header)

    @patch(JSON_SEND_FUNC)
    @patch(RAND_FUNC, _get_random_name)
    def test_sync_start_incorrect_interval(self, mock_send_api_req):
        mock_send_api_req.side_effect = [
            [{'name': 'RegionOne', 'syncStatus': '',
              'syncInterval': 0.0}],
            [{}],
            [{'syncStatus': 'syncInProgress',
              'requestId': self._get_random_name()}]
        ]
        assert self.drv.sync_start()
        calls = [
            ('region/RegionOne', 'PUT',
             [{'name': 'RegionOne',
               'syncInterval': self.drv.sync_interval}]),
            ('region/RegionOne/sync', 'POST',
             {'requester': socket.gethostname().split('.')[0],
              'requestId': self._get_random_name()})
        ]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    @patch(JSON_SEND_FUNC)
    @patch(RAND_FUNC, _get_random_name)
    def test_sync_end(self, mock_send_api_req):
        mock_send_api_req.return_value = [{'requester':
                                           self._get_random_name()}]
        self.drv.current_sync_name = self._get_random_name()
        self.assertTrue(self.drv.sync_end())
        calls = [
            ('region/RegionOne/sync', 'DELETE')
        ]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    @patch(JSON_SEND_FUNC)
    def test_create_region(self, mock_send_api_req):
        self.drv.create_region('foo')
        calls = [('region/', 'POST', [{'name': 'foo'}])]
        self._verify_send_api_request_call(mock_send_api_req, calls)

    @patch('requests.Response')
    def test_get_region_exception(self, mock_response):
        mock_response.ok.return_value = False
        self.assertIsNone(self.drv.get_region('foo'))

    @patch('requests.Response')
    def test_get_cvx_uuid_exception(self, mock_response):
        mock_response.ok.return_value = False
        self.assertIsNone(self.drv.get_cvx_uuid())



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from multiprocessing import Queue
import random

from neutron.agent import rpc as agent_rpc
from neutron_lib.agent import topics
from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context

from networking_arista.ml2 import arista_sync
from networking_arista.ml2.mechanism_arista import AristaDriver


class AristaHASimulationDriver(AristaDriver):

    def __init__(self):
        super(AristaHASimulationDriver, self).__init__()
        self.provision_queue1 = Queue()
        self.provision_queue2 = Queue()
        self.provision_queue3 = Queue()
        self.provision_queues = [self.provision_queue1,
                                 self.provision_queue2,
                                 self.provision_queue3]

    def get_workers(self):
        return [arista_sync.AristaSyncWorker(self.provision_queue1),
                arista_sync.AristaSyncWorker(self.provision_queue2),
                arista_sync.AristaSyncWorker(self.provision_queue3)]

    def create_network_postcommit(self, context):
        self.provision_queue = random.choice(self.provision_queues)
        super(AristaHASimulationDriver, self).create_network_postcommit(
            context)

    def update_network_postcommit(self, context):
        self.provision_queue = random.choice(self.provision_queues)
        super(AristaHASimulationDriver, self).update_network_postcommit(
            context)

    def delete_network_postcommit(self, context):
        self.provision_queue = random.choice(self.provision_queues)
        super(AristaHASimulationDriver, self).delete_network_postcommit(
            context)

    def update_port_postcommit(self, context):
        self.provision_queue = random.choice(self.provision_queues)
        super(AristaHASimulationDriver, self).update_port_postcommit(context)

    def delete_port_postcommit(self, context):
        self.provision_queue = random.choice(self.provision_queues)
        super(AristaHASimulationDriver, self).delete_port_postcommit(context)


class AristaHAScaleSimulationDriver(AristaHASimulationDriver):

    def __init__(self):
        super(AristaHAScaleSimulationDriver, self).__init__()
        self.context = None
        self.plugin_rpc = None

    def initialize(self):
        super(AristaHAScaleSimulationDriver, self).initialize()
        self.context = context.get_admin_context_without_session()
        # Subscribe to port updates to force ports to active after binding
        # since a fake virt driver is being used, so OVS will never see
        # the libvirt interfaces come up, triggering the OVS provisioning
        self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        registry.subscribe(self._port_update_callback,
                           resources.PORT, events.AFTER_UPDATE)

    def _port_update_callback(self, resource, event, trigger, **kwargs):
        port = kwargs.get('port')
        host = port.get(portbindings.HOST_ID)
        vif_type = port.get(portbindings.VIF_TYPE)
        device_dict = {'device': port['id'],
                       'agent_id': 'ovs-agent-%s' % host,
                       'host': host}
        if vif_type == 'ovs':
            self.plugin_rpc.update_device_up(self.context, **device_dict)
        elif (port.get(portbindings.VNIC_TYPE) == 'normal'
              and vif_type == 'unbound'):
            self.plugin_rpc.update_device_down(self.context, **device_dict)



# Copyright (c) 2016 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from mock import patch
from neutron_lib import context
from oslo_config import cfg

from neutron.tests.unit.plugins.ml2.drivers import test_helpers
from neutron.tests.unit import testlib_api

import networking_arista.common.exceptions as exc
from networking_arista.ml2.type_drivers.driver_helpers import VlanSyncService
from networking_arista.ml2.type_drivers.type_arista_vlan \
    import AristaVlanTypeDriver
import networking_arista.tests.unit.utils as utils


EAPI_SEND_FUNC = ('networking_arista.ml2.rpc.arista_eapi.AristaRPCWrapperEapi'
                  '._send_eapi_req')


class AristaTypeDriverTest(testlib_api.SqlTestCase):

    def setUp(self):
        super(AristaTypeDriverTest, self).setUp()
        utils.setup_arista_wrapper_config(cfg)

    @patch(EAPI_SEND_FUNC)
    def test_initialize_type_driver(self, mock_send_eapi_req):
        type_driver = AristaVlanTypeDriver()
        type_driver.sync_service._force_sync = False
        type_driver.sync_service._vlan_assignment_uuid = {'uuid': 1}
        type_driver.sync_service._rpc = mock.MagicMock()
        rpc = type_driver.sync_service._rpc
        rpc.get_vlan_assignment_uuid.return_value = {'uuid': 1}
        type_driver.initialize()

        cmds = ['show openstack agent uuid',
                'show openstack resource-pool vlan region RegionOne uuid']

        calls = [mock.call(cmds=[cmd], commands_to_log=[cmd])
                 for cmd in cmds]
        mock_send_eapi_req.assert_has_calls(calls)
        type_driver.timer.cancel()


class AristaTypeDriverHelpersTest(test_helpers.HelpersTest):

    def setUp(self):
        utils.setup_arista_wrapper_config(cfg)
        super(AristaTypeDriverHelpersTest, self).setUp()
        self.driver = AristaVlanTypeDriver()
        self.driver.network_vlan_ranges = test_helpers.NETWORK_VLAN_RANGES

    def test_allocate_specific_allocated_segment_outside_pools(self):
        # Invalid test case for Arista type driver because the first
        # allocate fails with VlanUnavailable
        pass

    def test_allocate_specific_unallocated_segment_outside_pools(self):
        expected = dict(physical_network=test_helpers.TENANT_NET,
                        vlan_id=test_helpers.VLAN_OUTSIDE)
        self.assertRaises(exc.VlanUnavailable,
                          self.driver.allocate_fully_specified_segment,
                          self.context, **expected)


class VlanSyncServiceTest(testlib_api.SqlTestCase):
    """Test that VLANs are synchronized between EOS and Neutron."""

    def setUp(self):
        super(VlanSyncServiceTest, self).setUp()
        self.rpc = mock.MagicMock()
        self.sync_service = VlanSyncService(self.rpc)
        self.ctx = context.get_admin_context()

    def tearDown(self):
        super(VlanSyncServiceTest, self).tearDown()
        # Cleanup the db
        utils.delete_vlan_allocation(self.ctx)

    def _ensure_in_db(self, assigned, allocated, available):
        vlans = utils.get_vlan_allocation(self.ctx)
        self.assertEqual(len(vlans), len(assigned))
        used_vlans = []
        available_vlans = []
        for vlan in vlans:
            self.assertIn(vlan.vlan_id, assigned)
            if vlan.vlan_id in available:
                self.assertFalse(vlan.allocated)
                available_vlans.append(vlan.vlan_id)
            elif vlan.vlan_id in allocated:
                self.assertTrue(vlan.allocated)
                used_vlans.append(vlan.vlan_id)
        self.assertEqual(set(used_vlans), set(allocated))
        self.assertEqual(set(available_vlans), set(available))

    def _get_vlan_allocations(self):
        vlan_allocations = {
            'available_vlans': [],
            'allocated_vlans': [],
        }
        vlans = utils.get_vlan_allocation(self.ctx)
        for vlan in vlans:
            if vlan.allocated:
                vlan_allocations['allocated_vlans'].append(vlan.vlan_id)
            else:
                vlan_allocations['available_vlans'].append(vlan.vlan_id)
        return vlan_allocations

    def test_synchronization_before_region_sync(self):
        """Test VLAN sync with empty data from CVX"""

        # Populated VlanAllocations before starting the sync
        for seg_id in range(2, 500):
            utils.create_vlan_allocation(self.ctx, segmentation_id=seg_id)

        self.rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '10, 100',
            'availableVlans': '',
            'allocatedVlans': ''
        }
        self.sync_service.synchronize()

        self.assertTrue(self.sync_service._force_sync)

        # Verify only assignedVlans exist in the db
        vlans = self._get_vlan_allocations()
        assigned_vlans = [10, 100]
        allocated_vlans = []
        self.assertEqual(set(vlans['available_vlans']), set(assigned_vlans))
        self.assertEqual(set(vlans['allocated_vlans']), set(allocated_vlans))

    def test_synchronization_test(self):
        """Test VLAN sync based on allocated VLANs in db and CVX"""

        # Add entries to vlan allocation table
        VLAN_MIN = 100
        VLAN_MAX = 300
        for seg_id in range(VLAN_MIN, VLAN_MAX + 1):
            allocated = seg_id in [VLAN_MIN, VLAN_MAX]
            utils.create_vlan_allocation(self.ctx, segmentation_id=seg_id,
                                         allocated=allocated)

        # Test case that vlan resource pool does not have allocated vlans
        self.rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '10-20, 50-60, %s, %s' % (VLAN_MIN, VLAN_MAX),
            'availableVlans': '10-20, 50-60',
            'allocatedVlans': ''
        }
        self.sync_service.synchronize()

        self.assertFalse(self.sync_service._force_sync)

        allocated_vlans = [VLAN_MIN, VLAN_MAX]
        available_vlans = list(set(range(10, 21)) | set(range(50, 61)))
        assigned_vlans = list(set(available_vlans) | set(allocated_vlans))
        self._ensure_in_db(assigned_vlans, allocated_vlans, available_vlans)

        # Test case that vlan resource pool has updated resources
        self.rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '200-220, %s, %s' % (VLAN_MIN, VLAN_MAX),
            'availableVlans': '200-220',
            'allocatedVlans': '%s, %s' % (VLAN_MIN, VLAN_MAX)
        }
        available_vlans = list(set(range(200, 221)))
        assigned_vlans = list(set(available_vlans) | set(allocated_vlans))
        self.sync_service.synchronize()
        self._ensure_in_db(assigned_vlans, allocated_vlans, available_vlans)

    def test_synchronization_test_with_data_from_cvx(self):
        """Test VLAN sync based on data from CVX"""

        self.rpc.get_vlan_allocation.return_value = {
            'assignedVlans': '51-60,71-80',
            'availableVlans': '51-55,71,73,75,77,79',
            'allocatedVlans': '56-60,72,74,76,78,80'
        }
        assigned_vlans = list(set(range(51, 61)) | set(range(71, 81)))
        available_vlans = [51, 52, 53, 54, 55, 71, 73, 75, 77, 79]
        allocated_vlans = list(set(assigned_vlans) - set(available_vlans))
        self.sync_service.synchronize()

        self.assertFalse(self.sync_service._force_sync)

        self._ensure_in_db(assigned_vlans, allocated_vlans, available_vlans)



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import socket

from oslo_log import log as logging
from oslo_utils import excutils
import requests
import six
from six.moves.urllib.parse import quote

from networking_arista._i18n import _, _LI, _LW
from networking_arista.common import exceptions as arista_exc
from networking_arista.ml2.rpc.base import AristaRPCWrapperBase

LOG = logging.getLogger(__name__)


class AristaRPCWrapperJSON(AristaRPCWrapperBase):
    def __init__(self):
        super(AristaRPCWrapperJSON, self).__init__()
        self.current_sync_name = None

    def _get_url(self, host="", user="", password=""):
        return ('https://%s:%s@%s/openstack/api/' %
                (user, password, host))

    def _api_host_url(self, host=""):
        return self._get_url(host, quote(self._api_username()),
                             quote(self._api_password()))

    def _send_request(self, host, path, method, data=None,
                      sanitized_data=None):
        request_headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'X-Sync-ID': self.current_sync_name
        }
        url = self._api_host_url(host=host) + path
        # Don't log the password
        log_url = self._get_url(host=host, user=self._api_username(),
                                password="*****") + path

        resp = None
        data = json.dumps(data)
        try:
            msg = (_('JSON request type: %(type)s url %(url)s data: '
                     '%(data)s sync_id: %(sync)s') %
                   {'type': method, 'url': log_url,
                    'data': sanitized_data or data,
                    'sync': self.current_sync_name})
            LOG.info(msg)
            func_lookup = {
                'GET': requests.get,
                'POST': requests.post,
                'PUT': requests.put,
                'PATCH': requests.patch,
                'DELETE': requests.delete
            }
            func = func_lookup.get(method)
            if not func:
                LOG.warning(_LW('Unrecognized HTTP method %s'), method)
                return None

            resp = func(url, timeout=self.conn_timeout, verify=False,
                        data=data, headers=request_headers)
            msg = (_LI('JSON response contains: %(code)s %(resp)s') %
                   {'code': resp.status_code,
                   'resp': resp.json()})
            LOG.info(msg)
            if resp.ok:
                return resp.json()
            else:
                raise arista_exc.AristaRpcError(msg=resp.json().get('error'))
        except requests.exceptions.ConnectionError:
            msg = (_('Error connecting to %(url)s') % {'url': url})
            LOG.warning(msg)
        except requests.exceptions.ConnectTimeout:
            msg = (_('Timed out connecting to API request to %(url)s') %
                   {'url': url})
            LOG.warning(msg)
        except requests.exceptions.Timeout:
            msg = (_('Timed out during API request to %(url)s') %
                   {'url': url})
            LOG.warning(msg)
        except requests.exceptions.InvalidURL:
            msg = (_('Ignore attempt to connect to invalid URL %(url)s') %
                   {'url': self._server_ip})
            LOG.warning(msg)
        except ValueError:
            LOG.warning(_LW("Ignoring invalid JSON response: %s"), resp.text)
        except Exception as error:
            msg = six.text_type(error)
            LOG.warning(msg)
            # reraise the exception
            with excutils.save_and_reraise_exception() as ctxt:
                ctxt.reraise = True
        return {} if method == 'GET' else None

    def _check_if_cvx_leader(self, host):
        url = 'agent/'
        data = self._send_request(host, url, 'GET')
        return False if not data else data.get('isLeader', False)

    def _get_eos_master(self):
        cvx = self._get_cvx_hosts()
        for self._server_ip in cvx:
            if self._check_if_cvx_leader(self._server_ip):
                return self._server_ip
        return None

    def send_api_request(self, path, method, data=None, sanitized_data=None):
        host = self._get_eos_master()
        if not host:
            msg = six.text_type("Could not find CVX leader")
            LOG.info(msg)
            self.set_cvx_unavailable()
            raise arista_exc.AristaRpcError(msg=msg)
        self.set_cvx_available()
        return self._send_request(host, path, method, data, sanitized_data)

    def _set_region_update_interval(self):
        path = 'region/%s' % self.region
        data = {
            'name': self.region,
            'syncInterval': self.sync_interval
        }
        self.send_api_request(path, 'PUT', [data])

    def register_with_eos(self, sync=False):
        self.create_region(self.region)
        self._set_region_update_interval()

    def get_cvx_uuid(self):
        path = 'agent/'
        try:
            data = self.send_api_request(path, 'GET')
            return data.get('uuid', None)
        except arista_exc.AristaRpcError:
            return None

    def create_region(self, region):
        path = 'region/'
        data = {'name': region}
        return self.send_api_request(path, 'POST', [data])

    def get_region(self, name):
        path = 'region/%s' % name
        try:
            regions = self.send_api_request(path, 'GET')
            for region in regions:
                if region['name'] == name:
                    return region
        except arista_exc.AristaRpcError:
            pass
        return None

    def sync_start(self):
        LOG.info('Attempt to start sync')
        self.current_sync_name = None
        try:
            region = self.get_region(self.region)

            # If the region doesn't exist, we may need to create
            # it in order for POSTs to the sync endpoint to succeed
            if not region:
                self.register_with_eos()
                return False

            if region.get('syncInterval') != self.sync_interval:
                self._set_region_update_interval()

            if region and region['syncStatus'] == 'syncInProgress':
                LOG.info('Sync in progress, not syncing')
                return False

            req_id = self._get_random_name()
            data = {
                'requester': socket.gethostname().split('.')[0],
                'requestId': req_id
            }
            path = 'region/' + self.region + '/sync'
            self.send_api_request(path, 'POST', data)
            self.current_sync_name = req_id
            return True
        except (KeyError, arista_exc.AristaRpcError):
            LOG.info('Not syncing due to RPC error')
            return False
        LOG.info('Not syncing due to server syncStatus')
        return False

    def sync_end(self):
        LOG.info('Attempting to end sync')
        try:
            path = 'region/' + self.region + '/sync'
            self.send_api_request(path, 'DELETE')
            self.current_sync_name = None
            return True
        except arista_exc.AristaRpcError:
            LOG.info('Not ending sync due to RPC error')
            return False



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import constants as n_const
from neutron_lib.services.trunk import constants as t_const
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


SUPPORTED_NETWORK_TYPES = [
    n_const.TYPE_VLAN,
    n_const.TYPE_VXLAN]


SUPPORTED_DEVICE_OWNERS = [
    n_const.DEVICE_OWNER_COMPUTE_PREFIX,
    n_const.DEVICE_OWNER_BAREMETAL_PREFIX,
    n_const.DEVICE_OWNER_DHCP,
    n_const.DEVICE_OWNER_DVR_INTERFACE,
    n_const.DEVICE_OWNER_ROUTER_HA_INTF,
    n_const.DEVICE_OWNER_ROUTER_INTF,
    t_const.TRUNK_SUBPORT_OWNER]


UNSUPPORTED_DEVICE_OWNERS = [
    n_const.DEVICE_OWNER_COMPUTE_PREFIX + 'probe']


UNSUPPORTED_DEVICE_IDS = [
    n_const.DEVICE_ID_RESERVED_DHCP_PORT]

SUPPORTED_SG_PROTOCOLS = [
    None,
    n_const.PROTO_NAME_TCP,
    n_const.PROTO_NAME_UDP,
    n_const.PROTO_NAME_ICMP]

LOG = logging.getLogger(__name__)


def supported_device_owner(device_owner):

    if (any([device_owner.startswith(supported_owner) for
             supported_owner in SUPPORTED_DEVICE_OWNERS]) and
        not any([device_owner.startswith(unsupported_owner) for
                 unsupported_owner in UNSUPPORTED_DEVICE_OWNERS])):
        return True

    LOG.debug('Unsupported device owner: %s', device_owner)
    return False


def hostname(hostname, extra=None):
    fqdns_used = cfg.CONF.ml2_arista['use_fqdn']
    return hostname if fqdns_used else hostname.split('.')[0]


def physnet(hostname):
    fqdns_used = cfg.CONF.ml2_arista['use_fqdn_physnet']
    return hostname if fqdns_used else hostname.split('.')[0]



# Copyright (c) 2018 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log

from neutron_lib.api.definitions import port as p_api
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import trunk as t_api
from neutron_lib.api.definitions import trunk_details as td_api
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.plugins import directory
from neutron_lib.services.trunk import constants as t_const

from neutron.services.trunk.drivers import base

LOG = log.getLogger(__name__)

NAME = 'arista'
SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OTHER,
)
SUPPORTED_SEGMENTATION_TYPES = (
    t_const.SEGMENTATION_TYPE_VLAN,
)


class AristaTrunkDriver(base.DriverBase):

    @property
    def is_loaded(self):
        try:
            return NAME in cfg.CONF.ml2.mechanism_drivers
        except cfg.NoSuchOptError:
            return False

    @registry.receives(resources.TRUNK_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, **kwargs):
        """Called in trunk plugin's AFTER_INIT"""
        super(AristaTrunkDriver, self).register(resource, event,
                                                trigger, kwargs)
        registry.subscribe(self.subport_create,
                           resources.SUBPORTS, events.AFTER_CREATE)
        registry.subscribe(self.subport_delete,
                           resources.SUBPORTS, events.AFTER_DELETE)
        registry.subscribe(self.trunk_create,
                           resources.TRUNK, events.AFTER_CREATE)
        registry.subscribe(self.trunk_update,
                           resources.TRUNK, events.AFTER_UPDATE)
        registry.subscribe(self.trunk_delete,
                           resources.TRUNK, events.AFTER_DELETE)
        self.core_plugin = directory.get_plugin()
        LOG.debug("Arista trunk driver initialized.")

    @classmethod
    def create(cls):
        return cls(NAME, SUPPORTED_INTERFACES, SUPPORTED_SEGMENTATION_TYPES,
                   can_trunk_bound_port=True)

    def bind_port(self, parent):
        ctx = context.get_admin_context()
        trunk = parent.get(td_api.TRUNK_DETAILS, {})
        subports = trunk.get(t_api.SUB_PORTS, [])
        subport_ids = [subport['port_id'] for subport in subports]
        self._bind_subports(ctx, subport_ids, parent)
        trunk_plugin = directory.get_plugin(t_api.ALIAS)
        trunk_plugin.update_trunk(ctx, trunk.get('trunk_id'),
                                  {t_api.TRUNK:
                                   {'status': t_const.TRUNK_ACTIVE_STATUS}})

    def _bind_subports(self, ctx, subport_ids, parent):
        host_id = parent.get(portbindings.HOST_ID)
        vnic_type = parent.get(portbindings.VNIC_TYPE)
        profile = parent.get(portbindings.PROFILE)
        device_id = parent.get('device_id')
        for subport_id in subport_ids:
            self.core_plugin.update_port(
                ctx, subport_id,
                {p_api.RESOURCE_NAME:
                 {portbindings.HOST_ID: host_id,
                  portbindings.VNIC_TYPE: vnic_type,
                  portbindings.PROFILE: profile,
                  'device_owner': t_const.TRUNK_SUBPORT_OWNER,
                  'device_id': device_id,
                  'status': n_const.PORT_STATUS_ACTIVE}})

    def _unbind_subports(self, ctx, subport_ids, parent):
        for subport_id in subport_ids:
            self.core_plugin.update_port(
                ctx, subport_id,
                {p_api.RESOURCE_NAME:
                 {portbindings.HOST_ID: None,
                  portbindings.VNIC_TYPE: None,
                  portbindings.PROFILE: None,
                  'device_owner': '',
                  'device_id': '',
                  'status': n_const.PORT_STATUS_DOWN}})

    def _delete_trunk(self, trunk):
        ctx = context.get_admin_context()
        parent_id = trunk.port_id
        parent = self.core_plugin.get_port(ctx, parent_id)
        if parent.get(portbindings.VNIC_TYPE) != portbindings.VNIC_BAREMETAL:
            return
        subport_ids = [subport.port_id
                       for subport in trunk.sub_ports]
        self._unbind_subports(ctx, subport_ids, parent)

    def trunk_create(self, resource, event, trunk_plugin, payload):
        ctx = context.get_admin_context()
        parent_id = payload.current_trunk.port_id
        parent = self.core_plugin.get_port(ctx, parent_id)
        if parent.get(portbindings.VNIC_TYPE) != portbindings.VNIC_BAREMETAL:
            return
        subport_ids = [subport.port_id
                       for subport in payload.current_trunk.sub_ports]
        self._bind_subports(ctx, subport_ids, parent)
        trunk_plugin.update_trunk(ctx, payload.trunk_id,
                                  {t_api.TRUNK:
                                   {'status': parent['status']}})

    def trunk_update(self, resource, event, trunk_plugin, payload):
        if payload.current_trunk.status != t_const.TRUNK_ACTIVE_STATUS:
            self._delete_trunk(payload.current_trunk)

    def trunk_delete(self, resource, event, trunk_plugin, payload):
        self._delete_trunk(payload.original_trunk)

    def subport_create(self, resource, event, trunk_plugin, payload):
        ctx = context.get_admin_context()
        parent_id = payload.current_trunk.port_id
        parent = self.core_plugin.get_port(ctx, parent_id)
        if parent.get(portbindings.VNIC_TYPE) != portbindings.VNIC_BAREMETAL:
            return
        subport_ids = [subport.port_id for subport in payload.subports]
        self._bind_subports(ctx, subport_ids, parent)
        trunk_plugin.update_trunk(ctx, payload.trunk_id,
                                  {t_api.TRUNK:
                                   {'status': parent['status']}})

    def subport_delete(self, resource, event, trunk_plugin, payload):
        ctx = context.get_admin_context()
        parent_id = payload.current_trunk.port_id
        parent = self.core_plugin.get_port(ctx, parent_id)
        if parent.get(portbindings.VNIC_TYPE) != portbindings.VNIC_BAREMETAL:
            return
        subport_ids = [subport.port_id for subport in payload.subports]
        self._unbind_subports(ctx, subport_ids, parent)
        trunk_plugin.update_trunk(ctx, payload.trunk_id,
                                  {t_api.TRUNK:
                                   {'status': parent['status']}})
